import abc
import difflib
import os
import re
from itertools import groupby
from pathlib import Path
from typing import Any, Dict, Generator, List, Mapping, Optional, Protocol, Tuple, Union

from annet import cli_args, filtering, patching, rulebook
from annet.annlib import jsontools
from annet.annlib.diff import (  # pylint: disable=unused-import
    colorize_line,
    diff_cmp,
    diff_ops,
    gen_pre_as_diff,
    resort_diff,
)
from annet.annlib.netdev.views.hardware import HardwareView
from annet.annlib.output import LABEL_NEW_PREFIX, format_file_diff
from annet.cli_args import ShowDiffOptions
from annet.connectors import CachedConnector
from annet.output import output_driver_connector
from annet.storage import Device
from annet.types import Diff, PCDiff, PCDiffFile
from annet.vendors import registry_connector, tabparser

from .gen import Loader, old_new


def _diff_files(hw, old_files, new_files):
    ret = {}
    differ = file_differ_connector.get()
    for (path, (new_text, reload_data)) in new_files.items():
        old_text = old_files.get(path)
        is_new = old_text is None
        diff_lines = differ.diff_file(hw, path, old_text, new_text)
        ret[path] = (diff_lines, reload_data, is_new)
    return ret


def pc_diff(hw, hostname: str, old_files: Dict[str, str], new_files: Dict[str, str]) -> Generator[
    PCDiffFile, None, None]:
    sorted_lines = sorted(_diff_files(hw, old_files, new_files).items())
    for (path, (diff_lines, _reload_data, is_new)) in sorted_lines:
        if not diff_lines:
            continue
        label = hostname + os.sep + path
        if is_new:
            label = LABEL_NEW_PREFIX + label
        yield PCDiffFile(label=label, diff_lines=diff_lines)


def json_fragment_diff(
    hw,
    hostname: str,
    old_files: Dict[str, Any],
    new_files: Dict[str, Tuple[Any, Optional[str]]],
) -> Generator[PCDiffFile, None, None]:
    def jsonify_multi(files):
        return {
            path: jsontools.format_json(cfg)
            for path, cfg in files.items()
        }

    def jsonify_multi_with_cmd(files):
        ret = {}
        for path, cfg_reload_cmd in files.items():
            cfg, reload_cmd = cfg_reload_cmd
            ret[path] = (jsontools.format_json(cfg), reload_cmd)
        return ret

    jold, jnew = jsonify_multi(old_files), jsonify_multi_with_cmd(new_files)
    return pc_diff(hw, hostname, jold, jnew)


def worker(
    device_id,
    args: cli_args.DiffOptions,
    stdin,
    loader: Loader,
    filterer: filtering.Filterer,
) -> Union[Diff, PCDiff, None]:
    for res in old_new(
        args,
        config=args.config,
        loader=loader,
        no_new=args.clear,
        do_files_download=True,
        device_ids=[device_id],
        filterer=filterer,
        stdin=stdin,
    ):
        old = res.get_old(args.acl_safe)
        new = res.get_new(args.acl_safe)
        device = res.device
        acl_rules = res.get_acl_rules(args.acl_safe)
        new_files = res.get_new_files(args.acl_safe)
        new_json_fragment_files = res.get_new_file_fragments(args.acl_safe)

        pc_diff_files = []
        if res.old_files or new_files:
            pc_diff_files.extend(pc_diff(res.device.hw, device.hostname, res.old_files, new_files))
        if res.old_json_fragment_files or new_json_fragment_files:
            pc_diff_files.extend(json_fragment_diff(res.device.hw, device.hostname, res.old_json_fragment_files,
                                                    new_json_fragment_files))

        if pc_diff_files:
            pc_diff_files.sort(key=lambda f: f.label)
            return PCDiff(hostname=device.hostname, diff_files=pc_diff_files)
        elif old is not None:
            orderer = patching.Orderer.from_hw(device.hw)
            rb = rulebook.get_rulebook(device.hw)
            diff_tree = patching.make_diff(
                old,
                orderer.order_config(new),
                rb,
                [acl_rules, res.filter_acl_rules],
            )
            diff_tree = patching.strip_unchanged(diff_tree)
            return diff_tree


def gen_sort_diff(
    diffs: Mapping[Device, Union[Diff, PCDiff]], args: ShowDiffOptions
) -> Generator[Tuple[str, Generator[str, None, None] | str, bool], None, None]:
    """
    Возвращает осортированный дифф, совместимый с write_output
    :param diffs: Маппинг устройства в дифф
    :param args: Параметры коммандной строки
    """
    if args.no_collapse:
        devices_to_diff = {(dev,): diff for dev, diff in diffs.items()}
    else:
        non_pc_diffs = {dev: diff for dev, diff in diffs.items() if not isinstance(diff, PCDiff)}
        devices_to_diff = collapse_diffs(non_pc_diffs)
        devices_to_diff.update({(dev,): diff for dev, diff in diffs.items() if isinstance(diff, PCDiff)})
    for devices, diff_obj in devices_to_diff.items():
        if not diff_obj:
            continue
        if isinstance(diff_obj, PCDiff):
            for diff_file in diff_obj.diff_files:
                diff_text = (
                    "\n".join(diff_file.diff_lines)
                    if args.no_color
                    else "\n".join(format_file_diff(diff_file.diff_lines))
                )
                yield diff_file.label, diff_text, False
        else:
            output_driver = output_driver_connector.get()
            dest_name = ", ".join([output_driver.cfg_file_names(dev)[0] for dev in devices])
            pd = patching.make_pre(resort_diff(diff_obj))
            yield dest_name, gen_pre_as_diff(pd, args.show_rules, args.indent, args.no_color), False


def _transform_text_diff_for_collapsing(text_diff) -> List[str]:
    for line_no, line in enumerate(text_diff):
        text_diff[line_no] = re.sub(r"(snmp-agent .+) cipher \S+ (.+)", r"\1 cipher ENCRYPTED \2", line)
    return text_diff


def _make_text_diff(device: Device, diff: Diff) -> List[str]:
    formatter = registry_connector.get().match(device.hw).make_formatter()
    res = formatter.diff(diff)
    return res


def collapse_diffs(diffs: Mapping[Device, Diff]) -> Dict[Tuple[Device, ...], Diff]:
    """
    Группировка диффов.
    :param diffs:
    :return: дикт аналогичный типу Diff, но с несколькими dev в ключе.
        Нужно учесть что дифы сверяются в отформатированном виде
    """
    diffs_with_test = {dev: [diff, _transform_text_diff_for_collapsing(_make_text_diff(dev, diff))] for dev, diff in
                       diffs.items()}
    res = {}
    for _, collapsed_diff_iter in groupby(sorted(diffs_with_test.items(), key=lambda x: (x[0].hw.vendor, x[1][1])),
                                          key=lambda x: x[1][1]):
        collapsed_diff = list(collapsed_diff_iter)
        res[tuple(x[0] for x in collapsed_diff)] = collapsed_diff[0][1][0]

    return res


class FileDiffer(Protocol):
    @abc.abstractmethod
    def diff_file(self, hw: HardwareView, path: str | Path, old: str, new: str) -> list[str]:
        raise NotImplementedError


class UnifiedFileDiffer(FileDiffer):
    def __init__(self):
        self.context: int = 3

    def diff_file(self, hw: HardwareView, path: str | Path, old: str, new: str) -> list[str]:
        """Calculate the differences for config files.

        Args:
            hw: device hardware info
            path: path to file on a device
            old (Optional[str]): The old file content.
            new (Optional[str]): The new file content.

        Returns:
            List[str]: List of difference lines.
        """
        return self._diff_text_file(old, new)

    def _diff_text_file(self, old, new):
        """Calculate the differences for plaintext files."""
        context = self.context
        old_lines = old.splitlines() if old else []
        new_lines = new.splitlines() if new else []
        context = max(len(old_lines), len(new_lines)) if context is None else context
        return list(difflib.unified_diff(old_lines, new_lines, n=context, lineterm=""))


class FrrFileDiffer(UnifiedFileDiffer):
    def diff_file(self, hw: HardwareView, path: str | Path, old: str, new: str) -> list[str]:
        if (hw.PC.Mellanox or hw.PC.NVIDIA) and (path == "/etc/frr/frr.conf"):
            return self._diff_frr_conf(hw, old, new)
        return super().diff_file(hw, path, old, new)

    def _diff_frr_conf(self, hw: HardwareView, old_text: str | None, new_text: str | None) -> list[str]:
        """Calculate the differences for frr.conf files."""
        indent = "  "
        rb = rulebook.rulebook_provider_connector.get()
        rulebook_data = rb.get_rulebook(hw)
        formatter = registry_connector.get().match(hw).make_formatter(indent=indent)

        old_tree = tabparser.parse_to_tree(old_text or "", splitter=formatter.split)
        new_tree = tabparser.parse_to_tree(new_text or "", splitter=formatter.split)

        diff_tree = patching.make_diff(old_tree, new_tree, rulebook_data, [])
        pre_diff = patching.make_pre(diff_tree)
        diff_iterator = gen_pre_as_diff(pre_diff, show_rules=False, indent=indent, no_color=True)

        return [line.rstrip() for line in diff_iterator if "frr version" not in line]


class _FileDifferConnector(CachedConnector[FileDiffer]):
    name = "Device file diff processor"
    ep_name = "file_differ"


file_differ_connector = _FileDifferConnector()



from ipaddress import IPv4Network, IPv6Network, ip_network
from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Literal

from annet.rpl import RoutingPolicy, PrefixMatchValue, OrLonger


class CommunityLogic(Enum):
    AND = "AND"
    OR = "OR"


class CommunityType(Enum):
    BASIC = "BASIC"
    RT = "RT"
    SOO = "SOO"
    COST = "COST"
    LARGE = "LARGE"


@dataclass(frozen=True)
class CommunityList:
    name: str
    members: Sequence[str]
    type: CommunityType = CommunityType.BASIC
    logic: CommunityLogic = CommunityLogic.OR
    use_regex: bool = False


@dataclass(frozen=True)
class RDFilter:
    name: str
    number: int
    members: Sequence[str]


@dataclass(frozen=True)
class AsPathFilter:
    name: str
    filters: Sequence[str]


@dataclass
class IpPrefixListMember:
    prefix: IPv4Network | IPv6Network
    or_longer: OrLonger = (None, None)

    def __post_init__(self):
        self.prefix = ip_network(self.prefix)


@dataclass
class IpPrefixList:
    name: str
    members: list[IpPrefixListMember]

    def __post_init__(self):
        for (i, m) in enumerate(self.members):
            if isinstance(m, str):
                self.members[i] = IpPrefixListMember(m)


def ip_prefix_list(
        name: str,
        members_or_str: Sequence[IpPrefixListMember | str],
        or_longer: OrLonger = (None, None),
) -> IpPrefixList:
    members: List[IpPrefixListMember] = []
    for m in members_or_str:
        if isinstance(m, str):
            m = IpPrefixListMember(
                prefix=ip_network(m),
                or_longer=or_longer,
            )
        elif m.or_longer == (None, None):
            m = IpPrefixListMember(
                prefix=m.prefix,
                or_longer=or_longer,
            )
        members.append(m)
    return IpPrefixList(
        name=name,
        members=members,
    )


def arista_well_known_community(community: str) -> str:
    if community == "65535:0":
        return "GSHUT"
    return community


def mangle_united_community_list_name(values: Sequence[str]) -> str:
    """Name for a list used as HAS_ANY between multiple lists"""
    return "_OR_".join(values)


class PrefixListNameGenerator:
    def __init__(self, prefix_lists: Sequence[IpPrefixList], policies: Sequence[RoutingPolicy]):
        self._prefix_lists = {x.name: x for x in prefix_lists}
        self._policies = {x.name: x for x in policies}  # this is here for a later use ~azryve@

    def get_prefix(self, name: str, match: PrefixMatchValue | None) -> IpPrefixList:
        orig_prefix = self._prefix_lists[name]
        override_name: Optional[str] = None
        override_orlonger: Optional[OrLonger] = None

        if match and match.or_longer != (None, None):
            ge, le = match.or_longer
            ge_str = "unset" if ge is None else str(ge)
            le_str = "unset" if le is None else str(le)
            override_name = f"{orig_prefix.name}_{ge_str}_{le_str}"
            override_orlonger = match.or_longer

        return IpPrefixList(
            name=override_name or name,
            members=[
                IpPrefixListMember(
                    x.prefix,
                    or_longer=override_orlonger or x.or_longer,
                )
                for x in orig_prefix.members
            ],
        )


def group_community_members(
    all_communities: dict[str, CommunityList], communities: list[str],
) -> dict[CommunityType, list[str]]:
    members: dict[CommunityType, list[str]] = defaultdict(list)
    for community_name in communities:
        community = all_communities[community_name]
        members[community.type].extend(community.members)
    return members


class JuniperPrefixListNameGenerator(PrefixListNameGenerator):
    def get_prefix(self, name: str, match: PrefixMatchValue | None) -> IpPrefixList:
        plist = super().get_prefix(name, match)
        flavour = self.get_plist_flavour(plist)
        # keep the orginal name for an orlonger match
        if flavour != "custom":
            plist.name = name
        return plist

    def get_type(self, name: str, match: PrefixMatchValue | None) -> Literal["prefix-list", "route-filter"]:
        orig_plist = self.get_prefix(name, None)
        plist = self.get_prefix(name, match)
        orig_flavour = self.get_plist_flavour(orig_plist)
        flavour = self.get_plist_flavour(plist)
        if orig_flavour == "custom" or flavour == "custom":
            return "route-filter"
        else:
            return "prefix-list"

    def get_plist_flavour(self, plist: IpPrefixList) -> Literal["simple", "orlonger", "custom"]:
        is_orlonger: bool = False
        is_custom: bool = False
        for m in plist.members:
            ge, le = m.or_longer

            # or_longer != (None, None)
            if ge is not None or le is not None:
                is_orlonger = True

            # orlonger != (n, None), where n is .../n
            if ge is not None and ge != m.prefix.prefixlen:
                is_custom = True
                break

            # orlonger != (None, n), where n is 32 or 128 (ipv4/ipv6)
            if le is not None and le != m.prefix.max_prefixlen:
                is_custom = True
                break

        if is_custom:
            return "custom"
        elif is_orlonger:
            return "orlonger"
        else:
            return "simple"



from collections import defaultdict
from dataclasses import dataclass
from typing import cast, List, Union, Iterable, Optional, TypedDict

from annet.storage import Query

FIELD_VALUE_SEPARATOR = ":"
ALLOWED_GLOB_GROUPS = ["site", "tag", "role", "device_type", "status", "tenant"]


class Filter(TypedDict, total=False):
    site: list[str]
    tag: list[str]
    role: list[str]
    name: list[str]
    device_type: list[str]
    status: list[str]
    tenant: list[str]


@dataclass
class NetboxQuery(Query):
    query: List[str]

    @classmethod
    def new(
            cls, query: Union[str, Iterable[str]],
            hosts_range: Optional[slice] = None,
    ) -> "NetboxQuery":
        if hosts_range is not None:
            raise ValueError("host_range is not supported")
        return cls(query=list(query))

    @property
    def globs(self):
        # We process every query host as a glob
        return self.query

    def parse_query(self) -> Filter:
        query_groups = parse_query(self.globs)
        return cast(Filter, query_groups)

    def is_empty(self) -> bool:
        return len(self.query) == 0

    def is_host_query(self) -> bool:
        if not self.globs:
            return False
        for q in self.globs:
            if FIELD_VALUE_SEPARATOR in q:
                return False
        return True


def parse_query(query: list[str]) -> dict[str, list[str]]:
    query_groups = defaultdict(list)
    for q in query:
        if FIELD_VALUE_SEPARATOR in q:
            glob_type, param = q.split(FIELD_VALUE_SEPARATOR, 2)
            if glob_type not in ALLOWED_GLOB_GROUPS:
                raise Exception(f"unknown query type: '{glob_type}'")
            if not param:
                raise Exception(f"empty param for '{glob_type}'")
            query_groups[glob_type].append(param)
        else:
            query_groups["name"].append(q)
    return dict(query_groups)



from dataclasses import dataclass

from annet.annlib.types import Op


@dataclass
class UserConfig:
    name: str
    privilege: int
    role: str
    secret_type: str
    secret: str


def _parse_user_config(config_line):
    """Convert a user config line into a dataclass. Config example:

    username someuser privilege 15 role network-admin secret sha512 $6$....

    privilege could be omitted if equal to 1
    role could be omitted
    secret could be omitted, 'nopassword' is provided instead
    """
    splstr = config_line.split()
    name = splstr[1]
    priv = 1
    role = ""
    secret_type = ""
    secret = ""
    if "privilege" in splstr:
        pos = splstr.index("privilege")
        priv = int(splstr[pos + 1])
    if "role" in splstr:
        pos = splstr.index("role")
        role = splstr[pos + 1]
    if "secret" in splstr:
        pos = splstr.index("secret")
        secret_type = splstr[pos + 1]
        secret = splstr[pos + 2]
    return UserConfig(name=name, privilege=priv, role=role, secret_type=secret_type, secret=secret)


def user(key, diff, **_):
    if diff[Op.ADDED] and not diff[Op.REMOVED]:
        for add in diff[Op.ADDED]:
            yield (True, add["row"], None)
    elif diff[Op.REMOVED] and not diff[Op.ADDED]:
        for _ in diff[Op.REMOVED]:
            yield (False, f"no username {key[0]}", None)
    else:
        for num, add in enumerate(diff[Op.ADDED]):
            new_user = _parse_user_config(add["row"])
            old_user = _parse_user_config(diff[Op.REMOVED][num]["row"])
            if new_user.privilege != old_user.privilege:
                yield (True, f"username {key[0]} privilege {new_user.privilege}", None)
            if new_user.role != old_user.role and not new_user.role:
                yield (True, f"no username {key[0]} role", None)
            elif new_user.role != old_user.role:
                yield (True, f"username {key[0]} role {new_user.role}", None)
            if new_user.secret != old_user.secret and not new_user.secret:
                yield (True, f"username {key[0]} nopassword", None)
            elif new_user.secret != old_user.secret:
                yield (True, f"username {key[0]} secret {new_user.secret_type} {new_user.secret}", None)



from itertools import groupby
from operator import itemgetter
from typing import Dict, List

from pygments import lex
from pygments.formatter import Formatter
from pygments.lexer import RegexLexer
from pygments.lexers import DiffLexer, YamlLexer  # pylint: disable=no-name-in-module
from pygments.token import Token

from .output import TextArgs


# =====
Warning = Token.Warning  # pylint: disable=redefined-builtin
Error = Token.Error

YAML_TERMINAL_COLORS = {
    Token.Literal.String: "green",
    Token.Literal.Scalar.Plain: "darkblue",
    Token.Punctuation.Indicator: "darkblue",
}

DIFF_TERMINAL_COLORS = {
    Token.Generic.Inserted: "green",
    Token.Generic.Deleted: "red",
    Token.Generic.Heading: "cyan_blue",
}

SWITCH_OUTPUT_TERMINAL_COLORS = {
    Warning: "yellow",
    Error: "red",
}


class SwitchOutputLexer(RegexLexer):
    name = "SwitchOutputLexer"
    aliases = ["switch_output_lexer"]
    tokens = {
        "root": [
            (r"[wW]arning.*\n", Warning),
            (r"Info.*\n", Warning),
            (r"[eE]rror.*\n", Error),
            (r".*\n", Token.Text),
        ]
    }


class CursesFormatter(Formatter):
    def __init__(self, **options):
        self.colorscheme = options.pop("scheme")
        Formatter.__init__(self, **options)

    def format(self, tokensource, outfile="") -> Dict[int, List[TextArgs]]:
        res = {}
        tokensource = list(tokensource)
        tmp_res = {}
        line_no = 0
        for ttype, values in groupby(tokensource, itemgetter(0)):
            color = self.colorscheme.get(ttype)
            for value in values:
                if line_no not in tmp_res:
                    tmp_res[line_no] = []
                if value[1].endswith("\n"):
                    if len(value[1]) > 1:
                        tmp_res[line_no].append([color, value[1].rstrip()])
                    line_no += 1
                else:
                    tmp_res[line_no].append([color, value[1]])

        for line_no, color_values in tmp_res.items():
            res[line_no] = []
            for color, values in groupby(color_values, itemgetter(0)):
                str_values = "".join([v[1] for v in values])
                res[line_no].append(TextArgs(str_values, color))
        return res


def format_yaml(txt):
    return CursesFormatter(scheme=YAML_TERMINAL_COLORS).format(lex(txt, YamlLexer()))


def format_diff(txt):
    return CursesFormatter(scheme=DIFF_TERMINAL_COLORS).format(lex(txt, DiffLexer()))


LEXERS = {
    "diff": (DiffLexer, DIFF_TERMINAL_COLORS),
    "yaml": (YamlLexer, YAML_TERMINAL_COLORS),
    "switch_out": (SwitchOutputLexer, SWITCH_OUTPUT_TERMINAL_COLORS),
}


def curses_format(txt, lexer) -> Dict[int, List[TextArgs]]:
    return CursesFormatter(scheme=LEXERS[lexer][1]).format(lex(txt, LEXERS[lexer][0]()))
